/*
   Bsp Loader Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com


   The purpose of this demo is to load a Quake 3 map in our application.
   This is done pretty easily by loading the data section by section.
   Once loaded this map can be rendered out to the scene.  This demo
   also has light maps which are added using simple multi-texture.

   This map was created using Q3Radiant.  This map is just a simple room
   with a few squares in it.  Q3 maps can be textured with either .tga
   images or .jpg (that is what the Q3 engine supports) so keep that in
   mind when trying to load maps others made.  If they use any .jpgs you
   either must convert to .tga or write code to load a .jpg.
*/


// Demo: Bsp Loader.


#define WIN32_LEAN_AND_MEAN// Trims down the libraries.
#define VC_LEANMEAN        // Trims even further.


#include<windows.h>
#include<gl/gl.h>
#include<gl/glu.h>
#include<gl/glext.h>
#include"CBspLevel.h"
#include"CCamera.h"


// Function prototypes...
bool InitializeGL();
void RenderScene();
void Shutdown();


// The first is a global device.
HDC g_HDC;


// Object to hold the level, list of textures, and a camera object.
CBspLevel BspLevel;
CTGALoader m_textures[100];
CTGALoader m_lightMaps[100];

CCamera Camera;

// Multitexture functions.
PFNGLACTIVETEXTUREARBPROC glActiveTextureARB = 0;
PFNGLMULTITEXCOORD2FARBPROC glMultiTexCoord2fARB = 0;


void SetupPixelFormat(HDC hDC)
{
   int nPixelFormat;

   static PIXELFORMATDESCRIPTOR pfd = {
         sizeof(PIXELFORMATDESCRIPTOR),   // size of structure.
         1,                               // always 1.
         PFD_DRAW_TO_WINDOW |             // support window
         PFD_SUPPORT_OPENGL |             // support OpenGl
         PFD_DOUBLEBUFFER,                // support double buffering
         PFD_TYPE_RGBA,                   // support RGBA
         16,                              // bit color mode
         0, 0, 0, 0, 0, 0,                // ignore color bits
         0,                               // no alpha buffer
         0,                               // ignore shift bit
         0,                               // no accumulation buffer
         0, 0, 0, 0,                      // ignore accumulation bits.
         16,                              // number of depth buffer bits.
         0,                               // number of stencil buffer bits.
         0,                               // 0 means no auxiliary buffer
         PFD_MAIN_PLANE,                  // The main drawing plane
         0,                               // this is reserved
         0, 0, 0 };                       // layer masks ignored.

   // This chooses the best pixel format and returns index.
   nPixelFormat = ChoosePixelFormat(hDC, &pfd);

   // This set pixel format to device context.
   SetPixelFormat(hDC, nPixelFormat, &pfd);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static HGLRC hRC;                         // Rendering context.
   static HDC hDC;                           // Device context.
   int width, height;                        // The window width and height.

   switch(message)
      {
         case WM_CREATE:                     // Windows creation.
            hDC = GetDC(hwnd);               // This gets the device context for our window.
            g_HDC = hDC;                     // Assigns the global device context to this one.
            SetupPixelFormat(hDC);           // Call the pixel format function.

            hRC = wglCreateContext(hDC);     // Creates the rendering context.
            wglMakeCurrent(hDC, hRC);        // Makes the rendering context.
            return 0;
            break;


         case WM_CLOSE:                      // Close message.
         case WM_DESTROY:
            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);           // Deletes the rendering context.

            PostQuitMessage(0);              // Says close the program.
            return 0;
            break;


         case WM_SIZE:                       // re-size message.
            height = HIWORD(lParam);         // This gets the height of the window.
            width = LOWORD(lParam);          // This gets the width of the window.

            if(height==0)                    // we don't want it to be possible for a
               {                             // height of 0.  If it is 0 me make it 1.
                  height = 1;
               }

            glViewport(0, 0, width, height); // resets the viewport to new dimensions.
            glMatrixMode(GL_PROJECTION);     // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            // calculate the aspect ratio of the window.
            gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 1000.0f);

            glMatrixMode(GL_MODELVIEW);      // Sets the projection matrix.
            glLoadIdentity();                // Reset the modelview matrix.

            return 0;
            break;


         case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);

                  default:
                     break;
               }
            break;


         default:
            break;
      }

   // What this does is pass all of the unhandled messages to DefWindowProc.
   return (DefWindowProc(hwnd, message, wParam, lParam));
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                        // A message variable.
   WNDCLASSEX windowClass;                         // Your Window class.
   HWND hwnd;                                      // The Window handle.
   bool isFinished;                                // True then exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);        // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;    // style of the window.
   windowClass.lpfnWndProc = WndProc;              // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                     // Extra class information.
   windowClass.cbWndExtra = 0;                     // Extra window information.
   windowClass.hInstance = hInstance;              // Handle of application Instance.
   // Handle of application Icon.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW);// mouse cursor
   windowClass.hbrBackground = NULL;               // background color.
   windowClass.lpszMenuName = NULL;                // name of the main menu.
   windowClass.lpszClassName = "UltimateGameProgrammingClass";// window class name.
   // icon when minimized.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // Create window...
   hwnd = CreateWindowEx(NULL,// The extended window style.
                        "UltimateGameProgrammingClass",// window Class name.
                        "Quake3 BSP - By the Programming Ace.",// window name.
                        WS_OVERLAPPEDWINDOW | WS_VISIBLE |// The window style.
                        WS_SYSMENU | WS_CLIPCHILDREN |// window style.
                        WS_CLIPSIBLINGS,// window style.
                        100, 100,// window x, y coordinate.
                        640, 480,// window width and height.
                        NULL,// handle to parent window.
                        NULL,// menu.
                        hInstance,// handle to app instance.
                        NULL);   // pointer to window creation data.
   
   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW); // This shows the window.
   UpdateWindow(hwnd);        // This forces a paint message.

   isFinished = false;        // False = running, True = not running.

   // If initialize fail (return false), then we don't want the program to run.
   if(!InitializeGL()) isFinished = true;

   // This is the messsage loop.
   while(!isFinished)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     isFinished = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

   // Shutdown (free) all resources we used up.
   Shutdown();

   return (int)msg.wParam;
}


bool InitializeGL()
{
   glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

   glEnable(GL_DEPTH_TEST);
   glEnable(GL_TEXTURE_2D);

   // With this we will load the list of extensions the hardware supports.
   char *extension = (char*)glGetString(GL_EXTENSIONS);

   // Check for multitexture support.
   if(strstr(extension, "GL_ARB_multitexture") == 0) return false;

   // Load the multitexture funtion and test if all went well.
   glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
	glMultiTexCoord2fARB	= (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");

   // Error checking to make sure functions loaded.
   if(!glActiveTextureARB || !glMultiTexCoord2fARB) return false;


   // Load the .bsp level.
   if(!BspLevel.LoadLevel("maps//ugpTestMap.bsp")) return false;
   BspLevel.SwapAxis();

   // Load the textures we will need.
   for(int i = 0; i < BspLevel.m_totalTextures; i++) 
	   {
	      if(strcmp("noshader", BspLevel.m_textures[i].file) == 0) continue;
         strcat(BspLevel.m_textures[i].file, ".tga");

         if(!m_textures[i].LoadTGA(BspLevel.m_textures[i].file))
            return false;

         glGenTextures(1, &m_textures[i].ID);
         glBindTexture(GL_TEXTURE_2D, m_textures[i].ID);

         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

         int type = m_textures[i].type == 3 ? GL_RGB : GL_RGBA;
         gluBuild2DMipmaps(GL_TEXTURE_2D, type, m_textures[i].imageWidth,
               m_textures[i].imageHeight, type, GL_UNSIGNED_BYTE,
               m_textures[i].image);
	   }

	// Load the lightmaps we will need.
   for(int i = 0; i < BspLevel.m_totalLightMaps; i++) 
	   {
	      ApplyGamma(&BspLevel.m_lightMaps[i].lightMap[0][0][0],
	                 128 * 128 * 3, 3, 10);

         glGenTextures(1, &m_lightMaps[i].ID);
         glBindTexture(GL_TEXTURE_2D, m_lightMaps[i].ID);

         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

         gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 128, 128, GL_RGB, GL_UNSIGNED_BYTE,
                           BspLevel.m_lightMaps[i].lightMap);
	   }

   // Set the camera's default position.
   Camera.SetCamera(0, 150, -150, 0, 150, -149, 0, 1, 0);

   return true;
}

void RenderScene()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Clears the screen.
   glLoadIdentity();                   //Reset modelview matrix for new frame.

   // Set the camera.
   gluLookAt(Camera.mPos.x, Camera.mPos.y, Camera.mPos.z,
             Camera.mView.x, Camera.mView.y, Camera.mView.z,
             Camera.mUp.x, Camera.mUp.y, Camera.mUp.z);


   for(int i = 0; i < BspLevel.m_totalFaces; i++)
      {
         if(BspLevel.m_faces[i].type != 1) continue;
         stQ3Face *pFace = &BspLevel.m_faces[i];

         glActiveTextureARB(GL_TEXTURE0_ARB);
         glEnable(GL_TEXTURE_2D);
		   glBindTexture(GL_TEXTURE_2D, m_textures[pFace->texID].ID);

		   glActiveTextureARB(GL_TEXTURE1_ARB);
         glEnable(GL_TEXTURE_2D);
		   glBindTexture(GL_TEXTURE_2D, m_lightMaps[pFace->lightmapID].ID);

		   glActiveTextureARB(GL_TEXTURE0_ARB);

         glBegin(GL_TRIANGLE_FAN);

         for(int j = pFace->startVertexIndex; j <
             pFace->startVertexIndex + pFace->totalVertices; j++)
            {
               glTexCoord2f(BspLevel.m_vertices[j].texCoord.tu,
                            BspLevel.m_vertices[j].texCoord.tv);
               glMultiTexCoord2fARB(GL_TEXTURE1_ARB,
                                    BspLevel.m_vertices[j].lightmapCoord.tu,
                                    BspLevel.m_vertices[j].lightmapCoord.tv);
               glVertex3f(BspLevel.m_vertices[j].vertex.x,
                          BspLevel.m_vertices[j].vertex.y,
                          BspLevel.m_vertices[j].vertex.z);
            }

         glEnd();
      }

   // Checking for input.  We move the camera based on the button(s) that were pressed.
   if(GetKeyState(VK_UP) & 0x80) Camera.MoveCamera(FORWARD);
   if(GetKeyState(VK_DOWN) & 0x80) Camera.MoveCamera(BACKWARD);
   if(GetKeyState(VK_LEFT) & 0x80) Camera.StrafeCam(STRAFE_LEFT);
   if(GetKeyState(VK_RIGHT) & 0x80) Camera.StrafeCam(STRAFE_RIGHT);

   // Rotate camera
   POINT mousePos;
	GetCursorPos(&mousePos);

   Camera.RotateByMouse(mousePos.x, mousePos.y, (640 >> 1), (480 >> 1));
   Camera.RotateByMouse(mousePos.x, mousePos.y, (640 >> 1), (480 >> 1));

	SetCursorPos((640 >> 1), (480 >> 1));

   SwapBuffers(g_HDC);
}


void Shutdown()
{
   // Release all resources.
   for(int i = 0; i < BspLevel.m_totalTextures; i++)
      {
         glDeleteTextures(1, &m_textures[i].ID);
         m_textures[i].FreeImage();
      }

   for(int i = 0; i < BspLevel.m_totalLightMaps; i++)
      {
         glDeleteTextures(1, &m_lightMaps[i].ID);
         m_lightMaps[i].FreeImage();
      }

   BspLevel.Shutdown();
}


// Recap:

// Loading and rendering a Quake 3 level is pretty easy.  Loading is nothing
// more than reading in each section of data into its allocated array then
// drawing everything out each frame.  In the next demo we'll add a PVS system.


// Copyright December 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com